/******************************************************************************
 This file is a part od FDD3000 Emulator

 Copyright (C) 2013,2014 by Sławomir Szczyrba <sszczyrba@gmail.com>
 Copyright (C) 2014 by Stefano Bodrato <stefano_bodrato hotmail dt com>

 This program is free software: you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License as published by the Free
 Software Foundation, either version 2 of the License, or (at your option) any
 later version.

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this program. If not, see <http://www.gnu.org/licenses/lgpl.html>
******************************************************************************/

//#define DEBUG 1
#undef DEBUG

#ifdef DEBUG
    #define MSG(format, ...) fprintf (stderr, format, ## __VA_ARGS__)
#else
    #define MSG(...)
#endif


#include <config.h>
#include <libspectrum.h>

#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/stat.h>        /* For mode constants */
#include <sys/types.h>
#include <unistd.h>


#if defined (__MINGW32__) || defined (_WIN32) || defined (_WIN32_) || defined (__WIN32__)

#include <windows.h>
#include <tchar.h>

#else

#include <semaphore.h>
#ifndef SEM_FAILED
#define SEM_FAILED -1
#endif
#include <sys/mman.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/socket.h>

#endif



/* IPC */

typedef struct tFdd2TiLink {
    uint8_t  fdd2ti;
    uint8_t  ti2fdd;
    uint8_t  fddAlive;
    uint8_t  tiAlive;
} fdd2TiLink;

fdd2TiLink * fddlink;

uint8_t data;
int 	ipc;
int 	connected;
char   *path;

#if defined (__MINGW32__) || defined (_WIN32) || defined (_WIN32_) || defined (__WIN32__)

HANDLE fdd2tiF, fdd2tiE, ti2fddF, ti2fddE, fddAlive;
HANDLE shm, shmf;
DWORD  dwRetVal = 0;

TCHAR *shmId = _T("\\fdd.data");
TCHAR szTempFileName[MAX_PATH];
TCHAR lpTempPathBuffer[MAX_PATH];

//char shmId[512];
TCHAR fdd2tiE_Id[]=TEXT("Global\\fdd.fdd2ti.empty");
TCHAR fdd2tiF_Id[]=TEXT("Global\\fdd.fdd2ti.full");
TCHAR ti2fddE_Id[]=TEXT("Global\\fdd.ti2fdd.empty");
TCHAR ti2fddF_Id[]=TEXT("Global\\fdd.ti2fdd.full");
TCHAR fddAlive_Id[]=TEXT("Global\\fdd.alive");

#else

int    shm;
sem_t *fdd2tiF,*fdd2tiE,*ti2fddF,*ti2fddE,*fddAlive;
const char * shmId       = "fdd.data";
const char * fddAlive_Id = "fdd.alive";
const char * fdd2tiE_Id  = "fdd.fdd2ti.empty";
const char * fdd2tiF_Id  = "fdd.fdd2ti.full";
const char * ti2fddE_Id  = "fdd.ti2fdd.empty";
const char * ti2fddF_Id  = "fdd.ti2fdd.full";

#endif

uint8_t fddRead() {
    uint8_t data=0;
#if defined (__MINGW32__) || defined (_WIN32) || defined (_WIN32_) || defined (__WIN32__)
    if (WaitForSingleObject(fdd2tiF,INFINITE) == 0 ) {
        data=fddlink->fdd2ti;
        ReleaseSemaphore(fdd2tiE,1,NULL);
    } else { connected = 0; }
#else
    sem_wait(fdd2tiF);
    data=fddlink->fdd2ti;
    sem_post(fdd2tiE);
#endif
    return data;
}

void fddWrite(int8_t d) {
#if defined (__MINGW32__) || defined (_WIN32) || defined (_WIN32_) || defined (__WIN32__)
    if(WaitForSingleObject(ti2fddE,INFINITE) == 0 ) {
        fddlink->ti2fdd=d;
        ReleaseSemaphore(ti2fddF,1,NULL);
    } else { connected = 0; }
#else
    sem_wait(ti2fddE);
    fddlink->ti2fdd=d;
    sem_post(ti2fddF);
#endif
}

uint8_t fddIsAlive() {
    uint8_t ok;
#if defined (__MINGW32__) || defined (_WIN32) || defined (_WIN32_) || defined (__WIN32__)
    WaitForSingleObject(fddAlive,INFINITE);
    ok=fddlink->fddAlive&fddlink->tiAlive;
    ReleaseSemaphore(fddAlive,1,NULL);
#else
    sem_wait(fddAlive);
    ok=fddlink->fddAlive&fddlink->tiAlive;
    sem_post(fddAlive);
#endif
    return ok;
}

void cleanUp() {
    int i;

    connected=0;
    if(!ipc) { return; }
    ipc=0;

#if defined (__MINGW32__) || defined (_WIN32) || defined (_WIN32_) || defined (__WIN32__)

    WaitForSingleObject(fddAlive,INFINITE);
    fddlink->tiAlive=0;
    i=fddlink->fddAlive;
    ReleaseSemaphore(fddAlive,1,NULL);

//    semResetAll();
    CloseHandle(fddAlive);
    CloseHandle(ti2fddF);
    CloseHandle(ti2fddE);
    CloseHandle(fdd2tiF);
    CloseHandle(fdd2tiE);

    UnmapViewOfFile(fddlink);
    if(!i) {
		CloseHandle(shm);
		CloseHandle(shmf);
    }

#else

    sem_wait(fddAlive);
    fddlink->tiAlive=0;
    i=fddlink->fddAlive;
    sem_post(fddAlive);

    sem_close(fddAlive);
    sem_close(ti2fddF);
    sem_close(ti2fddE);
    sem_close(fdd2tiF);
    sem_close(fdd2tiE);

    munmap(fddlink,sizeof(fdd2TiLink));
    if(!i) {
        sem_unlink(fdd2tiE_Id);
        sem_unlink(fdd2tiF_Id);
        sem_unlink(ti2fddE_Id);
        sem_unlink(ti2fddF_Id);
        sem_unlink(fddAlive_Id);
        shm_unlink(shmId);
    }

#endif

}

int openShm() {
    int trunc;
    int i;

    if(!ipc) {
#if defined (__MINGW32__) || defined (_WIN32) || defined (_WIN32_) || defined (__WIN32__)
    char * tmp;

//    tmp=getenv("TMP");
//    if(tmp==NULL) tmp=getenv("TEMP");
//    if(tmp==NULL) tmp=".";

//    strcpy(shmId,tmp);
//    strcat(shmId,"\\fdd.data");
    dwRetVal = GetTempPath(MAX_PATH, lpTempPathBuffer); // buffer for path
    if (dwRetVal > MAX_PATH || (dwRetVal == 0)) {
        MSG("tim397_openShm: GetTempPath failed\n");
        return -5;
    }
    snprintf(szTempFileName,MAX_PATH,"%s%s",lpTempPathBuffer,shmId);

    MSG("TMP Shm file : %s\n",szTempFileName);

	shmf = CreateFile(
        szTempFileName,
        GENERIC_READ|GENERIC_WRITE,
        FILE_SHARE_READ|FILE_SHARE_WRITE,
        NULL,
        OPEN_ALWAYS,
        FILE_ATTRIBUTE_NORMAL,
        NULL
    );

    if (shmf == NULL) {
        MSG("failed to open file %s. ErrorCode: %d\n",shmId,GetLastError());
		return -1;
	}

    shm = CreateFileMapping(
		shmf,                    // use paging file
		NULL,                    // default security
		PAGE_READWRITE,          // read/write access
		0,                       // maximum object size (high-order DWORD)
		256,                     // maximum object size (low-order DWORD)
		NULL
    );

	if(shm==NULL){
        MSG("failed to map file. ErrorCode: %d\n",GetLastError());
		CloseHandle(shmf);
		return -2;
	}
        //fddlink=(fdd2TiLink*)mmap(0,sizeof(fdd2TiLink),PROT_READ|PROT_WRITE,MAP_SHARED,shm,0);
    fddlink = (fdd2TiLink*)MapViewOfFile(shm,   // handle to map object
		FILE_MAP_ALL_ACCESS, // read/write permission
		0,
		0,
		sizeof(fdd2TiLink)
    );

	if(fddlink==NULL){
		MSG("tim397_openShm: MapViewOfFile (mmap) failed\n");
        CloseHandle(shm);
		CloseHandle(shmf);
		return -3;
    }

    fdd2tiE=OpenSemaphore(SEMAPHORE_MODIFY_STATE|SYNCHRONIZE,TRUE,fdd2tiE_Id);
    if(fdd2tiE==NULL){MSG("failed to create sem fdd2tie. ErrorCode: %d\n",GetLastError()); }

    fdd2tiF=OpenSemaphore(SEMAPHORE_MODIFY_STATE|SYNCHRONIZE,TRUE,fdd2tiF_Id);
    if(fdd2tiF==NULL){MSG("failed to create sem fdd2tif. ErrorCode: %d\n",GetLastError()); }

    ti2fddE=OpenSemaphore(SEMAPHORE_MODIFY_STATE|SYNCHRONIZE,TRUE,ti2fddE_Id);
    if(ti2fddE==NULL){MSG("failed to create sem ti2fdde. ErrorCode: %d\n",GetLastError()); }

    ti2fddF=OpenSemaphore(SEMAPHORE_MODIFY_STATE|SYNCHRONIZE,TRUE,ti2fddF_Id);
    if(ti2fddF==NULL){MSG("failed to create sem ti2fddf. ErrorCode: %d\n",GetLastError()); }

    fddAlive=OpenSemaphore(SEMAPHORE_MODIFY_STATE|SYNCHRONIZE,TRUE,fddAlive_Id);
    if(fddAlive==NULL){MSG("failed to create sem sync. ErrorCode: %d\n",GetLastError()); }

    if ((fdd2tiE==NULL) || (fdd2tiF==NULL) || (ti2fddE==NULL) || (ti2fddF==NULL) || (fddAlive==NULL)) {
		MSG("tim397_openShm: semaphore creation failed\n");
        return -4;
    }

#else

        shm = shm_open(shmId, O_RDWR, 0666);
        if (-1==shm) {
	    MSG("tim397_openShm: shm_open file %s failed\n",shmId);
	    return -1;
	}
        trunc=ftruncate(shm,sizeof(fdd2TiLink));
        if (-1==trunc) {
	    MSG("tim397_openShm: ftruncate failed\n");
	    return -2;
	}
        fddlink=(fdd2TiLink*)mmap(0,sizeof(fdd2TiLink),PROT_READ|PROT_WRITE,MAP_SHARED,shm,0);
        if(fddlink==MAP_FAILED) {
	    MSG("tim397_openShm: mmap failed\n");
	    return -3;
	}
        close(shm);

        fdd2tiE=sem_open(fdd2tiE_Id,O_RDWR,0666,1);
        if( SEM_FAILED == fdd2tiE) { MSG("fdd2tiE open failed\n"); }

        fdd2tiF=sem_open(fdd2tiF_Id,O_RDWR,0666,0);
        if( SEM_FAILED == fdd2tiF) { MSG("fdd2tiF open failed\n"); }

        ti2fddE=sem_open(ti2fddE_Id,O_RDWR,0666,1);
        if( SEM_FAILED == ti2fddE) { MSG("ti2fddE open failed\n"); }

        ti2fddF=sem_open(ti2fddF_Id,O_RDWR,0666,0);
        if( SEM_FAILED == ti2fddF) { MSG("ti2fddF open failed\n"); }

        fddAlive=sem_open(fddAlive_Id,O_RDWR,0666,1);
        if( SEM_FAILED == fddAlive){ MSG("fddAlive open failed\n"); }

	if ((fdd2tiE==SEM_FAILED) || (fdd2tiF==SEM_FAILED) || (ti2fddE==SEM_FAILED) || (ti2fddF==SEM_FAILED) || (fddAlive==SEM_FAILED)) {
	    MSG("tim397_openShm: semaphore creation failed\n");
	    return -4;
	}
#endif
        ipc=1;
    }
    if(ipc&&(!connected)) {
#if defined (__MINGW32__) || defined (_WIN32) || defined (_WIN32_) || defined (__WIN32__)
		WaitForSingleObject(fddAlive,INFINITE);
        if(fddlink->fddAlive){ fddlink->tiAlive=1; connected=1; }
		ReleaseSemaphore(fddAlive,1,NULL);
#else
        sem_wait(fddAlive);
        if(fddlink->fddAlive){ fddlink->tiAlive=1; connected=1; }
        sem_post(fddAlive);
#endif
    }
	return (0);
}


void closeShm(void) {
    if(ipc&&connected) {
#if defined (__MINGW32__) || defined (_WIN32) || defined (_WIN32_) || defined (__WIN32__)
	WaitForSingleObject(fddAlive,INFINITE);
	fddlink->tiAlive=0;
    ReleaseSemaphore(fddAlive,1,NULL);
#else
	sem_wait(fddAlive);
        fddlink->tiAlive=0;
        sem_post(fddAlive);
#endif
    }
}
